/******************************************************************************\
 * Technische Universitaet Darmstadt, Institut fuer Nachrichtentechnik
 * Copyright(c) 2004
 *
 * Author(s):
 *	Volker Fischer, Julian Cable, Oliver Haffenden
 *
 * Description:
 *  
 * This is an implementation of the CPacketSocket interface that wraps up a QT socket.
 *	
 *
 ******************************************************************************
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or(at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
\******************************************************************************/

#include "PacketSocketQT.h"
#include <qstringlist.h>
#include <qtimer.h>
#include <iostream>
#include <sstream>
#include <errno.h>

CPacketSocketQT::CPacketSocketQT():
	pPacketSink(NULL), HostAddrOut(), iHostPortOut(-1),
	SocketDevice(QSocketDevice:: Datagram /* UDP */ ),
	pSocketNotivRead(NULL)
{
	/* Generate the socket notifier and connect the signal */
	pSocketNotivRead = new QSocketNotifier(SocketDevice.socket(),
											QSocketNotifier::Read);

	/* Connect the "activated" signal */
	QObject::connect(pSocketNotivRead, SIGNAL(activated(int)),
					  this, SLOT(OnDataReceived()));
					  
	/* allow connection when others are listening */
	SocketDevice.setAddressReusable(true);
}

CPacketSocketQT::~CPacketSocketQT()
{
	if(pSocketNotivRead)
	{
		pSocketNotivRead->disconnect();
		delete pSocketNotivRead;
	}
}

// Set the sink which will receive the packets
void
CPacketSocketQT::SetPacketSink(CPacketSink * pSink)
{
	pPacketSink = pSink;
}

// Stop sending packets to the sink
void
CPacketSocketQT::ResetPacketSink(void)
{
	pPacketSink = NULL;
}

// Send packet to the socket
void
CPacketSocketQT::SendPacket(const vector < _BYTE > &vecbydata, uint32_t addr, uint16_t port)
{
	int bytes_written;
	/* Send packet to network */
	//cout << "CPacketSocketQT::SendPacket(" << vecbydata.size() << " bytes, " << addr << ", " << port << ") " << HostAddrOut.toString() << ":" << iHostPortOut << endl;
	if(addr==0)
		bytes_written = SocketDevice.writeBlock((char*)&vecbydata[0], vecbydata.size(), HostAddrOut, iHostPortOut);
	else
		bytes_written = SocketDevice.writeBlock((char*)&vecbydata[0], vecbydata.size(), QHostAddress(addr), port);
	/* should we throw an exception or silently accept? */
	/* the most likely cause is that we are sending unicast and no-one
	   is listening, or the interface is down, there is no route */
	if(bytes_written == -1)
	{
		QSocketDevice::Error x = SocketDevice.error();
		if(x != QSocketDevice::NetworkFailure)
			qDebug("error sending packet");
	}
}

_BOOLEAN
CPacketSocketQT::SetDestination(const string & strNewAddr)
{
	/* syntax
	   1:  <port>                send to port on localhost
	   2:  <ip>:<port>           send to port on host or port on m/c group
	   3:  <ip>:<ip>:<port>      send to port on m/c group via interface
	 */
	/* Init return flag and copy string in QT-String "QString" */
	int ttl = 127;
	_BOOLEAN bAddressOK = TRUE;
	QStringList parts = QStringList::split(":", strNewAddr.c_str(), TRUE);
	switch(parts.count())
	{
	case 1:
		bAddressOK = HostAddrOut.setAddress("127.0.0.1");
		iHostPortOut = parts[0].toUInt();
		break;
	case 2:
		bAddressOK = HostAddrOut.setAddress(parts[0]);
		iHostPortOut = parts[1].toUInt();
    	if(setsockopt(SocketDevice.socket(), IPPROTO_IP, IP_TTL,
				(char*)&ttl, sizeof(ttl))==SOCKET_ERROR)
			bAddressOK = FALSE;
		break;
	case 3:
		{
			QHostAddress AddrInterface;
			AddrInterface.setAddress(parts[0]);
			bAddressOK = HostAddrOut.setAddress(parts[1]);
			iHostPortOut = parts[2].toUInt();
			const SOCKET s = SocketDevice.socket();
#if QT_VERSION < 0x030000
			uint32_t mc_if = htonl(AddrInterface.ip4Addr());
#else
			uint32_t mc_if = htonl(AddrInterface.toIPv4Address());
#endif
			if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_IF,
						(char *) &mc_if, sizeof(mc_if)) == SOCKET_ERROR)
				bAddressOK = FALSE;
    		if(setsockopt(s, IPPROTO_IP, IP_MULTICAST_TTL,
						(char*) &ttl, sizeof(ttl)) == SOCKET_ERROR)
				bAddressOK = FALSE;
		}
		break;
	default:
		bAddressOK = FALSE;
	}
	return bAddressOK;
}


_BOOLEAN
CPacketSocketQT::GetDestination(string & str)
{
	stringstream s;
	s << HostAddrOut.toString().latin1() << ":" << iHostPortOut;
	str = s.str();
	return TRUE;
}


_BOOLEAN
CPacketSocketQT::SetOrigin(const string & strNewAddr)
{
	/* syntax
	   1:  <port>
	   2:  <group ip>:<port>
	   3:  <interface ip>:<group ip>:<port>
	   4:  <interface ip>::<port>
	   5:  :<group ip>:<port>
	 */
	int iPort=-1;
	QHostAddress AddrGroup, AddrInterface;
	QStringList parts = QStringList::split(":", strNewAddr.c_str(), TRUE);
	bool ok=true;
	switch(parts.count())
	{
	case 1:
		iPort = parts[0].toUInt(&ok);
		break;
	case 2:
		iPort = parts[1].toUInt(&ok);
		ok &= AddrGroup.setAddress(parts[0]);
		break;
	case 3:
		iPort = parts[2].toUInt(&ok);
		if(parts[0].length() > 0)
			ok &= AddrInterface.setAddress(parts[0]);
		if(parts[1].length() > 0)
			ok &= AddrGroup.setAddress(parts[1]);
		break;
	default:
		ok = false;
	}

	if(ok ==false)
	{
		return FALSE;
	}

	/* Multicast ? */

#if QT_VERSION < 0x030000
	uint32_t gp = AddrGroup.ip4Addr();
#else
	uint32_t gp = AddrGroup.toIPv4Address();
#endif
	if(gp == 0)
	{
		/* Initialize the listening socket. */
		SocketDevice.bind(AddrInterface, iPort);
	}
	else if((gp & 0xe0000000) == 0xe0000000)	/* multicast! */
	{
		struct ip_mreq mreq;

		/* Initialize the listening socket. Host address is 0 -> "INADDR_ANY" */
		bool ok = SocketDevice.bind(QHostAddress(UINT32(0)), iPort);
		if(ok == false)
		{
			//QSocketDevice::Error x = SocketDevice.error();
			throw CGenErr("Can't bind to port to receive packets");
		}

#if QT_VERSION < 0x030000
		mreq.imr_multiaddr.s_addr = htonl(AddrGroup.ip4Addr());
		mreq.imr_interface.s_addr = htonl(AddrInterface.ip4Addr());
#else
		mreq.imr_multiaddr.s_addr = htonl(AddrGroup.toIPv4Address());
		mreq.imr_interface.s_addr = htonl(AddrInterface.toIPv4Address());
#endif
		const SOCKET s = SocketDevice.socket();
		int n = setsockopt(s, IPPROTO_IP, IP_ADD_MEMBERSHIP,(char *) &mreq,
							sizeof(mreq));
		if(n == SOCKET_ERROR)
		{
			throw
				CGenErr(string
						("Can't join multicast group to receive packets: ") +
						 strerror(errno));
		}
	}
	else /* one address specified, but not multicast - listen on a specific interface */
	{
		/* Initialize the listening socket. */
		SocketDevice.bind(AddrGroup, iPort);
	}
	return TRUE;
}

void
CPacketSocketQT::OnDataReceived()
{
	vector < _BYTE > vecbydata(MAX_SIZE_BYTES_NETW_BUF);

	/* Read block from network interface */
	const int iNumBytesRead = SocketDevice.readBlock((char *) &vecbydata[0], MAX_SIZE_BYTES_NETW_BUF);

	if(iNumBytesRead > 0)
	{
		/* Decode the incoming packet */
		if(pPacketSink != NULL)
		{
			vecbydata.resize(iNumBytesRead);
#if QT_VERSION < 0x030000
			uint32_t addr = SocketDevice.peerAddress().ip4Addr();
#else
			uint32_t addr = SocketDevice.peerAddress().toIPv4Address();
#endif
			pPacketSink->SendPacket(vecbydata, addr, SocketDevice.peerPort());
		}
	}
}
